import asyncio
import logging
import os
import subprocess
import uuid
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from contextlib import contextmanager
from typing import Callable

process_pool = ProcessPoolExecutor()
thread_pool = ThreadPoolExecutor(thread_name_prefix='run.py thread_pool')
running_processes = []  # NOTE is used in rosys.test.Runtime to advance time slower until computation is done
log = logging.getLogger('rosys.run')


async def io_bound(callback: Callable, *args: any):
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(thread_pool, callback, *args)


async def cpu_bound(callback: Callable, *args: any):
    with cpu():
        try:
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(process_pool, callback, *args)
        except RuntimeError as e:
            if 'cannot schedule new futures after shutdown' not in str(e):
                raise


@contextmanager
def cpu():
    id = str(uuid.uuid4())
    running_processes.append(id)
    try:
        yield
    finally:
        running_processes.remove(id)


async def sh(command: list[str], timeout: float = 1) -> str:
    '''executes a shell command
    command: a sequence of program arguments as subprocess.Popen requires
    returns: stdout
    '''
    def run() -> str:
        try:
            result = subprocess.run(
                command,
                preexec_fn=os.setsid,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.STDOUT
            )
        except subprocess.TimeoutExpired:
            log.warning(f'{" ".join(command)} took longer than {timeout} s. Aborting.')
            return ''
        return result.stdout.decode('utf-8')
    #self.log.debug('executing sh command: ' + ' '.join(command))
    return await io_bound(run)



import logging
from typing import Any, Optional

import cv2
import numpy as np
from pydantic import BaseModel

from .image import ImageSize
from .point import Point
from .point3d import Point3d
from .rotation import Rotation


class Intrinsics(BaseModel):
    matrix: list[list[float]]
    distortion: list[float]
    rotation: Rotation
    size: ImageSize


log = logging.getLogger('rosys.world.calibration')


class Extrinsics(BaseModel):
    tilt: Optional[Rotation]
    yaw: float = 0.0
    translation: list[float] = [0.0, 0.0, 1.0]


class Calibration(BaseModel):
    intrinsics: Intrinsics
    extrinsics: Extrinsics = Extrinsics()

    @property
    def rotation(self) -> Rotation:
        tilt = self.extrinsics.tilt or Rotation.zero()
        return Rotation.from_euler(0, 0, self.extrinsics.yaw) * tilt * self.intrinsics.rotation

    @property
    def rotation_array(self) -> Any:
        tilt = self.extrinsics.tilt or Rotation.zero()
        yaw = self.extrinsics.yaw
        return np.dot([[np.cos(yaw), -np.sin(yaw), 0], [np.sin(yaw), np.cos(yaw), 0], [0, 0, 1]], tilt.R) @ self.intrinsics.rotation.R

    @property
    def is_complete(self) -> bool:
        return self.extrinsics.tilt is not None

    def project_to_image(self, world_point: Point3d) -> Point:
        world_points = np.array([world_point.tuple])
        R = self.rotation_array
        Rod = cv2.Rodrigues(R.T)[0]
        t = -R.T @ self.extrinsics.translation
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion)
        image_points, _ = cv2.projectPoints(world_points, Rod, t, K, D)
        return Point(x=image_points[0, 0, 0], y=image_points[0, 0, 1])

    def project_array_to_image(self, world_points: Any) -> Any:
        R = self.rotation_array
        Rod = cv2.Rodrigues(R.T)[0]
        t = -R.T @ self.extrinsics.translation
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion)
        image_points, _ = cv2.projectPoints(world_points, Rod, t, K, D)
        return image_points.reshape(-1, 2)

    def project_from_image(self, image_point: Point, target_height: float = 0) -> Optional[Point3d]:
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion)
        image_points = np.array(image_point.tuple, dtype=np.float32).reshape(1, 1, 2)
        image_points_ = cv2.undistortPoints(image_points, K, D).reshape(-1, 2)
        image_points__ = cv2.convertPointsToHomogeneous(image_points_).reshape(-1, 3)
        objPoints = image_points__ @ self.rotation_array.T
        Z = self.extrinsics.translation[-1]
        t = np.array(self.extrinsics.translation)
        floor_points = t.T - objPoints * (Z - target_height) / objPoints[:, 2:]

        reprojection = self.project_to_image(Point3d(x=floor_points[0, 0], y=floor_points[0, 1], z=target_height))
        if objPoints[0, -1] * np.sign(Z) > 0 or reprojection.distance(image_point) > 2:
            log.warning(f'reprojection failed with {reprojection.distance(image_point)} px')
            return None

        return Point3d(x=floor_points[0, 0], y=floor_points[0, 1], z=target_height)

    def project_array_from_image(self, image_points: Any, target_height: float = 0) -> Any:
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion)
        image_points_ = cv2.undistortPoints(image_points.astype(np.float32), K, D).reshape(-1, 2)
        image_points__ = cv2.convertPointsToHomogeneous(image_points_).reshape(-1, 3)
        objPoints = image_points__ @ self.rotation_array.T
        Z = self.extrinsics.translation[-1]
        t = np.array(self.extrinsics.translation)
        floor_points = t.T - objPoints * (Z - target_height) / objPoints[:, 2:]

        reprojection = self.project_array_to_image(floor_points)
        sign = objPoints[:, -1] * np.sign(Z)
        distance = np.linalg.norm(reprojection - image_points, axis=1)
        floor_points[np.logical_or(sign > 0, distance > 2), :] = np.nan

        return floor_points

    @staticmethod
    def from_points(world_points: list[Point3d], image_points: list[Point], image_size: ImageSize, f0: float):
        world_points = np.array([p.tuple for p in world_points], dtype=np.float32).reshape(1, -1, 3)
        image_points = np.array([p.tuple for p in image_points], dtype=np.float32).reshape(1, -1, 2)
        K0 = np.array([[f0, 0, image_size.width / 2], [0, f0, image_size.height / 2], [0, 0, 1]], dtype=np.float32)
        flags = cv2.CALIB_USE_INTRINSIC_GUESS
        _, K, D, rvecs, tvecs = cv2.calibrateCamera(world_points, image_points, image_size.tuple, K0, None, flags=flags)

        rotation0 = Rotation(R=np.eye(3).tolist())
        intrinsics = Intrinsics(matrix=K.tolist(), distortion=D.tolist()[0], rotation=rotation0, size=image_size)

        rotation = Rotation.from_rvec(rvecs[0]).T
        translation = (-np.array(rotation.R).dot(tvecs)).flatten().tolist()
        extrinsics = Extrinsics(tilt=rotation, translation=translation)

        return Calibration(intrinsics=intrinsics, extrinsics=extrinsics)



import numpy as np

from ..hardware import Hardware
from ..helpers import eliminate_2pi, sleep
from ..world import PathSegment, Point, Spline, World
from .spline import drive_spline, throttle


async def drive_path(world: World, hardware: Hardware, path: list[PathSegment]):
    for segment in path:
        await drive_spline(segment.spline, world, hardware, throttle_at_end=segment == path[-1], flip_hook=segment.backward)


async def drive_to(world: World, hardware: Hardware, target: Point):
    if world.robot.parameters.minimum_turning_radius:
        await drive_circle(world, hardware, target)

    robot_pose = world.robot.prediction
    approach_spline = Spline(
        start=robot_pose.point,
        control1=robot_pose.point.interpolate(target, 1/3),
        control2=robot_pose.point.interpolate(target, 2/3),
        end=target,
    )
    await drive_spline(approach_spline, world, hardware)


async def drive_circle(world: World, hardware: Hardware, target: Point):
    while True:
        direction = world.robot.prediction.point.direction(target)
        angle = eliminate_2pi(direction - world.robot.prediction.yaw)
        if abs(angle) < np.deg2rad(5):
            break
        linear = 0.5
        sign = 1 if angle > 0 else -1
        angular = linear / world.robot.parameters.minimum_turning_radius * sign
        await hardware.drive(*throttle(world, linear, angular))
        await sleep(0.1)



from typing import Optional
import numpy as np
from ..world import Camera
from .actor import Actor


class CameraProjector(Actor):
    interval: float = 1.0

    async def step(self):
        for camera in self.world.cameras.values():
            if camera.calibration is None or not camera.calibration.is_complete:
                continue
            if camera.projection is not None:
                continue
            self.update_projection(camera)

    @staticmethod
    def update_projection(camera: Camera, rows: int = 12, columns: int = 16):
        c, r = np.meshgrid(np.linspace(0, camera.calibration.intrinsics.size.width, columns),
                           np.linspace(0, camera.calibration.intrinsics.size.height, rows))
        image_points = np.stack((c.flatten(), r.flatten()), axis=1)
        floor_points = camera.calibration.project_array_from_image(image_points).reshape(rows, columns, 3)
        invalid = np.isnan(floor_points).any(axis=2)
        camera.projection = [
            [
                None if invalid[r, c] else [point[0], point[1]]
                for c, point in enumerate(row)
            ]
            for r, row in enumerate(floor_points)
        ]

    @staticmethod
    def allclose(array1: list[list[Optional[list[float]]]],
                 array2: list[list[Optional[list[float]]]]) -> bool:
        is_none1 = [point is None for row in array1 for point in row]
        is_none2 = [point is None for row in array2 for point in row]
        if is_none1 != is_none2:
            return False
        values1 = [point for row in array1 for point in row if point is not None]
        values2 = [point for row in array2 for point in row if point is not None]
        return np.allclose(values1, values2)



from __future__ import annotations

import abc
from typing import Any, Optional

import numpy as np
from pydantic import BaseModel, Field

from .calibration import Calibration, Extrinsics, Intrinsics
from .image import Image, ImageSize
from .rotation import Rotation


class Camera(BaseModel, abc.ABC):
    id: str
    calibration: Optional[Calibration] = None
    calibration_simulation: Optional[Calibration] = Field(None, description='only needed for simulation')
    projection: Optional[list[list[Optional[list[float]]]]] = Field(None, exclude=True)
    images: list[Image] = Field([], exclude=True)

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    @property
    def latest_image_uri(self) -> str:
        image = self.latest_captured_image
        if image is None:
            return 'camera/placeholder'
        return f'camera/{self.id}/{image.time}'

    @property
    def captured_images(self) -> list[Image]:
        return [i for i in self.images if i.data]

    @property
    def latest_captured_image(self) -> Optional[Image]:
        images = self.captured_images
        if images:
            return images[-1]

    def set_perfect_calibration(
        self,
        x: float = 0, y: float = 0, z: float = 1,
        yaw: float = 0, tilt_x: float = 0, tilt_y: float = 0,
        image_width=800, image_height=600,
    ) -> None:
        calibration = Calibration(
            intrinsics=Camera.create_intrinsics(image_width, image_height),
            extrinsics=Extrinsics(tilt=Rotation.from_euler(tilt_x, np.pi + tilt_y, 0), yaw=yaw, translation=[x, y, z]),
        )
        self.calibration_simulation = calibration
        self.calibration = calibration.copy(deep=True)

    @staticmethod
    def create_intrinsics(image_width: int = 800, image_height: int = 600) -> Intrinsics:
        c = 570
        size = ImageSize(width=image_width, height=image_height)
        K = [[c, 0, size.width / 2], [0, c, size.height / 2], [0, 0, 1]]
        D = [0, 0, 0, 0, 0]
        rotation = Rotation.zero()
        return Intrinsics(matrix=K, distortion=D, rotation=rotation, size=size)

    def __str__(self) -> str:
        return self.title



import asyncio
import time
from multiprocessing import Pipe
from typing import Any, Optional

from .. import run
from ..helpers import is_test
from ..world import PathSegment, Point, Pose, Spline
from . import Actor
from .pathplanning import (PlannerCommand, PlannerGrowMapCommand, PlannerObstacleDistanceCommand, PlannerProcess,
                           PlannerResponse, PlannerSearchCommand, PlannerTestCommand)


class PathPlanner(Actor):
    interval: float = 0.1

    def __init__(self):
        super().__init__()
        self.connection, process_connection = Pipe()
        self.process = PlannerProcess(process_connection, self.world.robot.shape.outline)
        self.responses: dict[str, Any] = {}

    async def startup(self):
        await super().startup()
        self.process.start()

    async def tear_down(self):
        await super().tear_down()
        if self.process.is_alive():
            self.process.kill()
        self.log.info(f'teardown of {self.process} completed ({self.process.is_alive})')
        self.connection.close()
        self.process.connection.close()

    async def step(self):
        if self.connection.poll():
            response = self.connection.recv()
            assert isinstance(response, PlannerResponse)
            if time.time() < response.deadline:
                self.responses[response.id] = response.content

    async def grow_map(self, points: list[Point], timeout: float = 3.0) -> None:
        return await self._call(PlannerGrowMapCommand(
            points=points,
            deadline=time.time()+timeout,
        ))

    async def search(self, *,
                     goal: Pose,
                     start: Optional[Pose] = None,
                     timeout: float = 3.0) -> list[PathSegment]:
        return await self._call(PlannerSearchCommand(
            areas=list(self.world.areas.values()),
            obstacles=list(self.world.obstacles.values()),
            start=start or self.world.robot.prediction,
            goal=goal,
            deadline=time.time()+timeout,
        ))

    async def test_spline(self, spline: Spline, timeout: float = 3.0) -> bool:
        return await self._call(PlannerTestCommand(
            areas=list(self.world.areas.values()),
            obstacles=list(self.world.obstacles.values()),
            spline=spline,
            deadline=time.time()+timeout,
        ))

    async def get_obstacle_distance(self, pose: Pose, timeout: float = 3.0) -> float:
        return await self._call(PlannerObstacleDistanceCommand(
            areas=list(self.world.areas.values()),
            obstacles=list(self.world.obstacles.values()),
            pose=pose,
            deadline=time.time()+timeout,
        ))

    async def _call(self, command: PlannerCommand) -> Any:
        with run.cpu():
            self.connection.send(command)
            while command.id not in self.responses:
                if time.time() > command.deadline:
                    raise TimeoutError(f'process call {command.id} did not respond in time')
                if is_test:
                    await self.step()  # NOTE: otherwise step() is not called while awaiting response
                await asyncio.sleep(self.interval)
            result = self.responses.pop(command.id)
            if isinstance(result, Exception):
                raise result
            return result



import json
import logging
import os

from .world import World


class Persistence:

    def __init__(self, world: World, filepath: str = '~/.rosys/world.json'):
        self.world = world
        self.filepath = os.path.expanduser(filepath)
        os.makedirs(os.path.dirname(self.filepath), exist_ok=True)
        self.log = logging.getLogger('rosys.Persistence')

    def dump(self) -> dict:
        robot_excludes = {
            'velocity',
            'shape',
            'parameters',
            'prediction',
            'detection',
            'simulation',
            'odometry',
            'current_velocity',
            'last_movement',
            'hardware_time',
            'clock_offset',
            'carrot',
        }
        return {
            'robot': self.world.robot.dict(exclude=robot_excludes),
            'areas': {id: area.dict() for id, area in self.world.areas.items()},
            'obstacles': {id: obstacle.dict() for id, obstacle in self.world.obstacles.items()},
            'usb_cameras': {id: camera.dict() for id, camera in self.world.usb_cameras.items()},
        }

    def load(self, world: World):
        self.world.robot = world.robot
        self.world.areas = world.areas
        self.world.obstacles = world.obstacles
        self.world.usb_cameras = world.usb_cameras

    def backup(self):
        os.makedirs(os.path.dirname(self.filepath), exist_ok=True)
        try:
            with open(self.filepath, 'w') as f:
                json.dump(self.dump(), f)
        except:
            self.log.exception('failed to backup')

    def parse_world(self, obj: dict) -> World:
        return self.world.parse_obj(obj)  # NOTE we use the provided world for parsing to load all attributes

    def restore(self):
        if not os.path.exists(self.filepath):
            self.log.warning('No backup file found.')
            return
        try:
            with open(self.filepath, 'r') as f:
                self.load(self.parse_world(json.load(f)))
        except:
            self.log.exception(f'Could not load from backup at {self.filepath}')



import inspect
from typing import Awaitable, Callable, Optional, Union

from nicegui.ui import Ui

from .. import Runtime
from .settings import update_interval


class AutomationControls:
    ui: Ui
    runtime: Runtime

    def __init__(self,
                 default_automation: Optional[Awaitable] = None,
                 can_start: Optional[Callable[[], Union[bool, Awaitable[bool]]]] = None) -> None:

        async def start():
            if inspect.iscoroutinefunction(can_start):
                if not await can_start():
                    return
            elif can_start is not None and not can_start():
                return
            self.runtime.automator.start(default_automation())

        def pause():
            self.runtime.automator.pause(because='pause button was pressed')

        def resume():
            self.runtime.automator.resume()

        def stop():
            self.runtime.automator.stop(because='stop button was pressed')

        play_button = self.ui.button(on_click=start).props('icon=play_arrow unelevated')
        pause_button = self.ui.button(on_click=pause).props('icon=pause outline')
        resume_button = self.ui.button(on_click=resume).props('icon=play_arrow outline')
        stop_button = self.ui.button(on_click=stop).props('icon=stop outline')

        def refresh():
            play_button.visible = self.runtime.automator.is_stopped
            pause_button.visible = self.runtime.automator.is_running
            resume_button.visible = self.runtime.automator.is_paused
            play_button.view.disable = default_automation is None or not self.runtime.automator.enabled
            stop_button.view.disable = self.runtime.automator.is_stopped
        self.ui.timer(update_interval, refresh)



import re
import shutil
from dataclasses import dataclass
from typing import Any, Optional

import cv2
import rosys
from numpy.typing import NDArray

from .. import event
from ..world import Image, ImageSize, UsbCamera
from .actor import Actor

MJPG = cv2.VideoWriter_fourcc(*'MJPG')
SCAN_INTERVAL = 10


@dataclass
class Device:
    '''devices specific infos are kept seperately (world should not be aware of them)'''
    uid: str
    video_id: int
    capture: Optional[Any] = None  # cv2.VideoCapture device
    exposure_min: int = 0
    exposure_max: int = 0
    exposure_default: int = 0


def process_image(image: NDArray, rotation: rosys.world.ImageRotation, crop: rosys.world.Rectangle = None) -> bytes:
    if crop is not None:
        image = image[int(crop.y):int(crop.y+crop.height), int(crop.x):int(crop.x+crop.width)]
    if rotation == rosys.world.ImageRotation.LEFT:
        image = cv2.rotate(image, cv2.ROTATE_90_COUNTERCLOCKWISE)
    elif rotation == rosys.world.ImageRotation.RIGHT:
        image = cv2.rotate(image, cv2.ROTATE_90_CLOCKWISE)
    elif rotation == rosys.world.ImageRotation.UPSIDE_DOWN:
        image = cv2.rotate(image, cv2.ROTATE_180)
    return cv2.imencode('.jpg', image)[1].tobytes()


class UsbCameraCapture(Actor):
    interval: float = 0.3

    def __init__(self):
        super().__init__()
        self.devices: dict[str, Device] = {}
        self.last_scan: Optional[float] = None

    async def step(self):
        await super().step()
        await self.update_device_list()
        for uid, camera in self.world.usb_cameras.items():
            if not camera.active:
                if uid in self.devices and self.devices[uid].capture is not None:
                    self.deactivate(camera)
                continue
            try:
                if uid not in self.devices or self.devices[uid].capture is None:
                    await self.activate(uid)
                if self.devices[uid].capture is None:
                    self.log.warn(f'unexpected missing capture handle for {uid}')
                    continue
                image = await rosys.run.io_bound(self.capture_image, uid)
                if image is None:
                    self.deactivate(camera)
                    continue
                bytes = await rosys.run.cpu_bound(process_image, image, camera.rotation, camera.crop)
                size = camera.resolution or ImageSize(width=800, height=600)
                camera.images.append(Image(camera_id=uid, data=bytes, time=self.world.time, size=size))
            except:
                self.log.exception(f'could not capture image from {uid}')
                self.deactivate(camera)
        self.purge_old_images()

    def capture_image(self, id) -> Any:
        _, image = self.devices[id].capture.read()
        return image

    async def activate(self, uid: str):
        camera = self.world.usb_cameras[uid]
        device = self.devices[uid]
        capture = self.get_capture_device(device.video_id)
        if capture is None:
            return
        self.devices[uid].capture = capture
        camera.fps = int(device.capture.get(cv2.CAP_PROP_FPS))
        if camera.exposure is None:
            camera.exposure = device.exposure_default / device.exposure_max
        await rosys.run.io_bound(self.set_parameters, camera, device)
        self.log.info(f'activated {uid}')

    def deactivate(self, camera: UsbCamera):
        if camera.id not in self.devices:
            return
        self.log.info(f'deactivated {camera.id}')
        self.devices[camera.id].capture.release()
        del self.devices[camera.id]

    def purge_old_images(self):
        for camera in self.world.usb_cameras.values():
            while camera.images and camera.images[0].time < self.world.time - 5 * 60.0:
                del camera.images[0]

    async def update_device_list(self):
        if self.last_scan is not None and self.world.time < self.last_scan + SCAN_INTERVAL:
            return
        self.last_scan = self.world.time
        output = await rosys.run.sh(['v4l2-ctl', '--list-devices'])
        output = '\n'.join([s for s in output.split('\n') if not s.startswith('Cannot open device')])
        for infos in output.split('\n\n'):
            match = re.search('\((.*)\)', infos)
            if match is None:
                continue
            uid = match.group(1)
            if uid not in self.world.usb_cameras:
                self.world.usb_cameras[uid] = UsbCamera(id=uid)
                self.log.info(f'adding camera {uid}')
                await event.call(event.Id.NEW_CAMERA, self.world.usb_cameras[uid])
            lines = infos.splitlines()
            if 'dev/video' not in lines[1]:
                continue
            num = int(lines[1].strip().lstrip('/dev/video'))
            if uid not in self.devices:
                self.devices[uid] = Device(uid=uid, video_id=num)
                await self.load_value_ranges(self.devices[uid])

    def get_capture_device(self, index: int):
        try:
            capture = cv2.VideoCapture(index)
            if capture is None:
                self.log.error(f'video device {index} is unavailable')
            elif not capture.isOpened():
                self.log.error(f'video device {index} can not be opened')
                capture.release()
            else:
                return capture
        except:
            self.log.exception(f'{index} device failed')

    async def tear_down(self):
        await super().tear_down()
        for camera in self.world.usb_cameras.values():
            self.deactivate(camera)
        self.devices.clear()

    @staticmethod
    def is_operable() -> bool:
        return shutil.which('v4l2-ctl') is not None

    def set_parameters(self, camera: UsbCamera, device: Device):
        # NOTE enforcing motion jpeg for now
        if device.capture.get(cv2.CAP_PROP_FOURCC) != MJPG:
            device.capture.set(cv2.CAP_PROP_FOURCC, MJPG)
        resolution = ImageSize(
            width=int(device.capture.get(cv2.CAP_PROP_FRAME_WIDTH)),
            height=int(device.capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
        )
        if camera.resolution and camera.resolution != resolution:
            self.log.info(f'updating resolution of {camera.id} from {resolution} to {camera.resolution}')
            device.capture.set(cv2.CAP_PROP_FRAME_WIDTH, camera.resolution.width)
            device.capture.set(cv2.CAP_PROP_FRAME_HEIGHT, camera.resolution.height)
        auto_exposure = device.capture.get(cv2.CAP_PROP_AUTO_EXPOSURE) == 3
        if camera.auto_exposure and not auto_exposure:
            self.log.info(f'activating auto-exposure for {camera.id}')
            device.capture.set(cv2.CAP_PROP_AUTO_EXPOSURE, 3)  # `v4l2-ctl -L` says "3: Aperture Priority Mode"
        if auto_exposure and not camera.auto_exposure:
            self.log.info(f'deactivating auto-exposure of {camera.id}')
            device.capture.set(cv2.CAP_PROP_AUTO_EXPOSURE, 1)  # `v4l2-ctl -L` says "1: Manual Mode"
        if not camera.auto_exposure:
            exposure = device.capture.get(cv2.CAP_PROP_EXPOSURE) / device.exposure_max
            if camera.exposure is not None and camera.exposure != exposure:
                self.log.info(f'updating exposure of {camera.id} from {exposure} to {camera.exposure})')
                device.capture.set(cv2.CAP_PROP_EXPOSURE, int(camera.exposure * device.exposure_max))

    async def load_value_ranges(self, device: Device) -> None:
        output = await self.run_v4l(device, '--all')
        match = re.search('exposure_absolute.*: min=(\d*).*max=(\d*).*default=(\d*).*', output)
        device.exposure_min = int(match.group(1))
        device.exposure_max = int(match.group(2))
        device.exposure_default = int(match.group(3))

    async def run_v4l(self, device: Device, *args):
        cmd = ['v4l2-ctl', '-d', str(device.video_id)]
        cmd.extend(args)
        return await rosys.run.sh(cmd)



from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field

from .image import Image


class UploadData(BaseModel):
    image: Image
    detector_name: str

    def __hash__(self) -> int:
        return hash(self.image)


class Upload(BaseModel):
    minimal_minutes_between_uploads: float = 1.0
    last_upload: datetime = datetime.fromtimestamp(0)
    queue: set[UploadData] = Field(set(), exclude=True)
    priority_queue: set[UploadData] = Field(set(), exclude=True)

    def mark(self, image: Image, *, force: bool = False, detector_name: Optional[str] = None):
        '''Mark an image for upload. Set force=True to circumvent minimal_minutes_between_uploads.'''
        data = UploadData(image=image, detector_name='detector' if detector_name is None else detector_name)
        if force:
            self.priority_queue.add(data)
        else:
            self.queue.add(data)

    def get_queued(self, detector_name: str) -> list[Image]:
        return [data.image for data in self.queue if data.detector_name == detector_name and data.image.data]

    def get_priority_queued(self, detector_name: str) -> list[Image]:
        return [data.image for data in self.priority_queue if data.detector_name == detector_name and data.image.data]



from ..communication import Communication
from ..world import World
from .hardware import Hardware


class CommunicatingHardware(Hardware):

    def __init__(self, world: World, communication: Communication) -> None:
        super().__init__(world)
        self.communication = communication

    @property
    def is_simulation(self) -> bool:
        return False



from __future__ import annotations

import asyncio
import logging
from asyncio.exceptions import CancelledError
from typing import Optional, Type, TypeVar

from . import Persistence, event, is_test, run, sleep, task_logger
from .actors import (Actor, AsyncioMonitor, Automator, Backup, CameraProjector, Detector, DetectorSimulator,
                     GarbageCollector, Lizard, NetworkMonitor, Odometer, PathPlanner, Steerer, UsbCameraCapture,
                     UsbCameraSimulator)
from .communication import CommunicationFactory
from .hardware import Hardware, RobotBrain, SimulatedHardware
from .world import World

T = TypeVar('T')


class Runtime:
    log = logging.getLogger(__name__)

    def __init__(self,
                 world: Optional[World] = None,
                 persistence: Optional[Persistence] = None,
                 hardware: Optional[Hardware] = None):
        self.world = world or World()
        Actor.world = self.world
        self.tasks = []
        if not is_test:
            self.persistence = persistence or Persistence(self.world)
            self.persistence.restore()

        self.hardware = hardware or self.create_hardware()
        if is_test:
            assert self.hardware.is_simulation, 'real hardware must not be used in tests'
        self.lizard = Lizard(self.hardware)
        self.odometer = Odometer()
        self.steerer = Steerer(self.hardware)
        self.automator = Automator()
        self.camera_projector = CameraProjector()
        self.asyncio_monitor = AsyncioMonitor()
        self.detector: Optional[Detector] = None  # NOTE can be set by runtime.with_detector()
        self.usb_camera_simulator: Optional[UsbCameraSimulator] = None  # NOTE can be set by runtime.with_usb_cameras()
        self.actors = [
            self.camera_projector,
            self.lizard,
            self.odometer,
            self.steerer,
            self.automator,
            self.asyncio_monitor,
            GarbageCollector(),
        ]
        if NetworkMonitor.is_operable():
            self.with_actors(NetworkMonitor())
        if not is_test:
            self.with_actors(Backup(self.persistence))
        self.path_planner: Optional[PathPlanner] = None

    def create_hardware(self):
        communication = CommunicationFactory.create()
        if communication is not None:
            return RobotBrain(self.world, communication)
        else:
            return SimulatedHardware(self.world)

    def with_actors(self, *actors: list[Actor]) -> Runtime:
        '''Adds list of additional actors to runtime.'''
        self.actors += actors
        return self

    def with_usb_cameras(self) -> Runtime:
        '''Adds USB camera capture actor to runtime.'''
        if UsbCameraCapture.is_operable() and not is_test:
            self.with_actors(UsbCameraCapture())
        else:
            self.usb_camera_simulator = UsbCameraSimulator()
            self.with_actors(self.usb_camera_simulator)
        return self

    def with_detector(self, real: Optional[Detector] = None, simulation: Optional[DetectorSimulator] = None) -> Runtime:
        '''Adds detector to runtime.'''
        if self.hardware.is_real and not is_test:
            self.detector = real or Detector()
        else:
            self.detector = simulation or DetectorSimulator()
        self.with_actors(self.detector)
        return self

    def with_path_planner(self) -> Runtime:
        '''Adds path planning actor to runtime.'''
        self.path_planner = PathPlanner()
        self.with_actors(self.path_planner)
        return self

    async def startup(self) -> None:
        if self.tasks:
            raise Exception('should be only executed once')

        event.register(event.Id.NEW_NOTIFICATION, self.store_notification)
        for actor in self.actors:
            try:
                await actor.startup()
            except:
                self.log.exception(f'error while starting {actor}')
                continue
            if actor.interval is not None:
                self.log.debug(f'starting actor {actor.name} with interval {actor.interval}s')
                self.tasks.append(task_logger.create_task(self.repeat(actor), name=actor.name))
        self.tasks.append(asyncio.create_task(self.watch_emitted_events(), name='watch_emitted_events'))

        if not is_test:
            await asyncio.sleep(1)  # NOTE we wait for RoSys to start up before analyzing async debugging
        self.activate_async_debugging()
        self.log.debug('startup completed')
        self.world.start_time = self.world.time

    async def shutdown(self) -> None:
        try:
            await self.hardware.drive(0, 0)
        except:
            pass
        if not is_test:
            self.persistence.backup()
        [t.cancel() for t in self.tasks]
        if not is_test:
            run.process_pool.shutdown()
        for a in self.actors:
            await a.tear_down()
        # await asyncio.gather(*[task_logger.create_task(a.tear_down(), name=f'{a.name}.tear_down()') for a in self.actors])

    async def repeat(self, actor: Actor):
        while True:
            start = self.world.time
            try:
                await actor.step()
                dt = self.world.time - start
            except (CancelledError, GeneratorExit):
                return
            except:
                dt = self.world.time - start
                self.log.exception(f'error in {actor}')
                if actor.interval == 0 and dt < 0.1:
                    delay = 0.1 - dt
                    self.log.warning(
                        f'{actor} would be called to frequently ' +
                        f'because it only took {dt*1000:.0f} ms; ' +
                        f'delaying this step for {delay*1000:.0f} ms')
                    await sleep(delay)
            try:
                await sleep(actor.interval - dt)
            except (CancelledError, GeneratorExit):
                return

    def get_actor(self, _type: Type[T]) -> T:
        return next((a for a in self.actors if type(a) is _type), None)

    def store_notification(self, message: str):
        self.log.info(message)
        self.world.notifications.append((self.world.time, message))

    def activate_async_debugging(self):
        '''Produce warnings for coroutines which take too long on the main loop and hence clog the event loop'''
        try:
            loop = asyncio.get_running_loop()
            loop.set_debug(True)
            loop.slow_callback_duration = 0.05
        except:
            self.log.exception('could not activate async debugging')

    async def watch_emitted_events(self):
        while True:
            try:
                for task in event.tasks:
                    if task.done() and task.exception():
                        self.handle_exception(task.exception())
                # cleanup finished tasks
                event.tasks = [t for t in event.tasks if not t.done()]
            except:
                self.log.exception('failing to watch emitted events')
            await sleep(0 if is_test else 0.1)

    def handle_exception(self, ex: Exception):
        self.log.exception('task failed to execute', exc_info=ex)
